// 002 u_c_ib_script
// 11 oct 2006

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_ib_script;
  interface
    uses Classes, u_c_basic_object
        , u_types_constants, u_c_text_file
        ;

    type t_symbol_type= (e_unknown_symbol,
                            e_identifier_symbol, e_integer_symbol, e_operator_symbol,
                            e_string_litteral_symbol,
                            e_comment_symbol,
                            e_end_symbol);

         c_ib_script= class(c_text_file)
                        m_symbol_type: t_symbol_type;
                        m_blank_string, m_symbol_string: String;
                        _m_previous_index: Integer;

                        m_c_result_list: tStringList;
                        m_request_index: Integer;

                        m_is_SET_TERM, m_is_COMMIT_WORK, m_is_SET_AUTODDL: Boolean;
                        m_terminator: Char;

                        Constructor create_ib_script(p_name, p_file_name: String);

                        function f_initialized: Boolean;
                        procedure initialize_scanner;
                        function f_read_symbol: Boolean;
                        procedure test_scanner;

                        function f_remove_comment: String;

                        procedure initialize_request_scanner;
                        function f_next_request: String;

                        function f_get_request: String;

                        Destructor Destroy; Override;
                      end; // c_ib_script

  implementation
    uses SysUtils, Dialogs, u_c_display;

    // -- c_ib_script

    Constructor c_ib_script.create_ib_script(p_name, p_file_name: String);
      begin
        Inherited create_text_file(p_name, p_file_name);
      end; // create_ib_script

    function c_ib_script.f_initialized: Boolean;
      begin
        if f_load
          then Result:= true
          else begin
              display('*** could not open '+ m_c_file_name.m_complete_name
                  + ' '+ IntToStr(m_io_result));
              Result:= False;
            end;
      end; // f_initialized

    procedure c_ib_script.initialize_scanner;
      begin
        _m_previous_index:= -1;
        m_buffer_index:= 0;
        m_terminator:= ';'
      end; // initialize_scanner

    function c_ib_script.f_read_symbol: Boolean;
        // -- true until eof
      const k_blanks= [' ', chr(9)];
            k_returns= [chr(13), chr(10)];
            k_blanks_returns= k_blanks+ k_returns;
            k_number= ['0'..'9', '.'];
            k_identifier= ['a'..'z', 'A'..'Z', '_', '0'..'9'];
            k_other= [#0..#255]- k_blanks_returns- k_identifier;

      procedure get_blanks;
        var l_start: Integer;
        begin
          l_start:= m_buffer_index;
          while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index] in k_blanks_returns) do
            inc(m_buffer_index);
          m_blank_string:= f_extract_string_start_end(l_start, m_buffer_index- 1);
        end; // get_blanks

      procedure get_identifier;
        var l_start: Integer;
        begin
          m_symbol_type:= e_identifier_symbol;
          l_start:= m_buffer_index;
          while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index] in k_identifier) do
            inc(m_buffer_index);

          m_symbol_string:= f_extract_string_start_end(l_start, m_buffer_index- 1);
        end; // get_identifier

      procedure get_number;
        var l_start: Integer;
        begin
          m_symbol_type:= e_integer_symbol;
          l_start:= m_buffer_index;
          while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index] in k_number) do
            inc(m_buffer_index);

          m_symbol_string:= f_extract_string_start_end(l_start, m_buffer_index- 1);
        end; // get_number

      procedure get_operator;
        var l_start: Integer;
        begin
          m_symbol_type:= e_operator_symbol;
          l_start:= m_buffer_index;
          inc(m_buffer_index);
          m_symbol_string:= f_extract_string_start_end(l_start, m_buffer_index- 1);
        end; // get_operator

      procedure get_other;
        var l_start: Integer;
        begin
          m_symbol_type:= e_operator_symbol;
          l_start:= m_buffer_index;
          inc(m_buffer_index);
          m_symbol_string:= f_extract_string_start_end(l_start, m_buffer_index- 1);
        end; // get_other

      procedure get_string_litteral(p_delimiter: Char);
        var l_start: Integer;
        begin
          m_symbol_type:= e_string_litteral_symbol;
          l_start:= m_buffer_index;

          // -- skip starting delimiter
          Inc(m_buffer_index);
          while (m_buffer_index< m_buffer_size) and (m_pt_buffer[m_buffer_index]<> p_delimiter) do
            Inc(m_buffer_index);

          // -- skip ending delimiter
          Inc(m_buffer_index);

          // -- ?? check -1
          m_symbol_string:= f_extract_string_start_end(l_start, m_buffer_index- 1);
        end; // get_string_litteral

      procedure skip_comment_slash_star;
        begin
          m_symbol_type:= e_comment_symbol;
          Inc(m_buffer_index, 2);

          while (m_buffer_index+ 1< m_buffer_size)
              and
                  (
                     (m_pt_buffer[m_buffer_index]<> '*')
                    or
                      (m_pt_buffer[m_buffer_index+ 1]<> '/')
                  )
                   do
            inc(m_buffer_index);

          Inc(m_buffer_index, 2);
        end; // skip_comment_slash_star

      begin // f_read_symbol
        m_symbol_string:= '';
        m_symbol_type:= e_unknown_symbol;

        if f_end_of_text
          then begin
              Result:= False;
              m_symbol_type:= e_end_symbol;
            end
          else begin
              get_blanks;

              if m_buffer_index< m_buffer_size
                then begin
                    case m_pt_buffer[m_buffer_index] of
                      'a'..'z', 'A'..'Z' : get_identifier;
                      '0'..'9' : get_number;
                      '''': get_string_litteral('''');
                      '"' : get_string_litteral('"');
                      '/' : if (m_buffer_index+ 1< m_buffer_size) and (m_pt_buffer[m_buffer_index+ 1]= '*')
                              then skip_comment_slash_star
                              else get_operator;
                      else get_other;
                    end; // case
                  end
                else m_symbol_type:= e_end_symbol;

              Result:= True;
            end; // not eof
      end; // f_read_symbol

    procedure c_ib_script.test_scanner;
      begin
        repeat
          f_read_symbol;
          display(m_symbol_string);
          if m_buffer_index= _m_previous_index
            then begin
                display_bug_stop('no_progress');
                Break;
              end;
          _m_previous_index:= m_buffer_index;

          if m_buffer_index> 3000
            then Break;
        until m_symbol_type= e_end_symbol;
      end; // test_scanner

    function c_ib_script.f_remove_comment: String;
      var l_previous_identifier: String;

      function f_is_SET_TERM: Boolean;
        begin
          Result:=   (l_previous_identifier= 'SET')
                    and
                     (m_symbol_string= 'TERM')
        end; // f_is_SET_TERM

      function f_is_COMMIT_WORK: Boolean;
        begin
          Result:=   (l_previous_identifier= 'COMMIT')
                    and
                     (m_symbol_string= 'WORK')
        end; // f_is_COMMIT_WORK

      function f_is_SET_AUTODDL: Boolean;
        begin
          Result:=   (l_previous_identifier= 'SET')
                    and
                     (m_symbol_string= 'AUTODDL')
        end; // f_is_SET_AUTODDL

      var l_terminator: Char;
          l_request: String;

      begin // f_remove_comment
        Result:= '';
        l_terminator:= ';';
        l_request:= '';
        m_c_result_list:=  tStringList.Create;

        repeat
          f_read_symbol;
          // display(Format('%7d ', [m_buffer_index])+ m_symbol_string);

          if m_symbol_type<> e_comment_symbol
            then begin
                if f_is_SET_TERM
                  then begin
                      // -- get the terminator
                      f_read_symbol;
                      l_terminator:= m_symbol_string[1];
                      display('// -- SET TERM '+ m_symbol_string);

                      // -- skip actual terminator
                      f_read_symbol;

                      // -- clear the request
                      l_request:= '';
                    end else
                if f_is_COMMIT_WORK
                  then begin
                      // -- skip ";"
                      f_read_symbol;
                      // -- clear the request
                      l_request:= '';
                    end else
                if f_is_SET_AUTODDL
                  then begin
                      // -- skip "OFF ;"
                      f_read_symbol;
                      f_read_symbol;
                      // -- clear the request
                      l_request:= '';
                    end else
                if m_symbol_string= l_terminator
                  then begin
                      display(l_request);
                      display('=====');

                      m_c_result_list.Add(Trim(l_request));
                      m_c_result_list.Add('==========');
                      l_request:= '';
                    end
                  else l_request:= l_request+ m_blank_string+ m_symbol_string;

                l_previous_identifier:= m_symbol_string;
                Result:= Result+ m_blank_string+ m_symbol_string;
              end;

          // display(m_symbol_string);

          if m_buffer_index= _m_previous_index
            then begin
                display_bug_stop('no_progress');
                Break;
              end;
          _m_previous_index:= m_buffer_index;

          if m_buffer_index> 500000
            then Break;
        until m_symbol_type= e_end_symbol;
      end; // f_remove_comment

    procedure c_ib_script.initialize_request_scanner;
      begin
        m_request_index:= 0;
      end; // initialize_request_scanner

    function c_ib_script.f_next_request: String;
      var l_the_line: String;
      begin
        Result:= '';
        with m_c_result_list do
        begin
          repeat
            if m_request_index>= Count
              then Break;
            l_the_line:= Strings[m_request_index];
          until False;
        end; // with m_c_result_list
      end; // f_next_request

    function c_ib_script.f_get_request: String;

      var l_request: String;

      procedure get_request;
        var l_previous_identifier: String;

        function f_is_SET_TERM: Boolean;
          begin
            Result:=   (l_previous_identifier= 'SET')
                      and
                       (m_symbol_string= 'TERM')
          end; // f_is_SET_TERM

        function f_is_COMMIT_WORK: Boolean;
          begin
            Result:=   (l_previous_identifier= 'COMMIT')
                      and
                       (m_symbol_string= 'WORK')
          end; // f_is_COMMIT_WORK

        function f_is_SET_AUTODDL: Boolean;
          begin
            Result:=   (l_previous_identifier= 'SET')
                      and
                       (m_symbol_string= 'AUTODDL')
          end; // f_is_SET_AUTODDL

        var l_display: Boolean;

        begin // get_request
          l_request:= '';
          l_display:= False;

          // -- break if found m_terminator
          repeat
            f_read_symbol;
            if l_display
              then display(Format('%7d ', [m_buffer_index])+ m_symbol_string);

            if m_symbol_type<> e_comment_symbol
              then begin
                  if f_is_SET_TERM
                    then begin
                        // -- get the terminator
                        f_read_symbol;
                        m_terminator:= m_symbol_string[1];
                        display('// -- SET TERM '+ m_symbol_string);
                        m_is_SET_TERM:= True;
                        // l_display:= True;

                        // -- skip actual terminator and exit loop
                        f_read_symbol;
                        break;
                      end else
                  if f_is_COMMIT_WORK
                    then begin
                        m_is_COMMIT_WORK:= True;
                      end else
                  if f_is_SET_AUTODDL
                    then begin
                        // -- skip "OFF ;"
                        f_read_symbol;
                        m_is_SET_AUTODDL:= True;
                      end else
                  if m_symbol_string= m_terminator
                    then begin
(*
                        display(l_request);
                        display('=====');
*)
                        Break;
                      end
                    else l_request:= l_request+ m_blank_string+ m_symbol_string;

                  l_previous_identifier:= m_symbol_string;
                end;

            // display(m_symbol_string);

            if m_buffer_index= _m_previous_index
              then begin
                  display_bug_stop('no_progress');
                  Break;
                end;
            _m_previous_index:= m_buffer_index;

            if m_buffer_index> 500000
              then begin
                  m_symbol_type:= e_end_symbol;
                  // Break;
                end;
          until m_symbol_type= e_end_symbol;

          if m_is_SET_TERM
            then ; // stop('');
        end; // get_request

      begin // f_get_request
        m_is_SET_TERM:= False;
        m_is_COMMIT_WORK:= False;
        m_is_SET_AUTODDL:= False;

        repeat
          get_request;
        until (m_symbol_type= e_end_symbol) or (m_symbol_type<> e_comment_symbol);
        if m_symbol_type= e_end_symbol
          then l_request:= '';
        Result:= l_request;
      end; // f_get_request

    Destructor c_ib_script.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_ib_script
    end. // u_c_ib_script


